import org.ejml.simple.SimpleMatrix;

import java.util.Random;

public class FCLayer extends Layer {
    SimpleMatrix weights;
    SimpleMatrix biases;

    public FCLayer(int inputSize, int outputSize) {
        Random random = new Random();
        weights = new SimpleMatrix(inputSize, outputSize, true,
                random.doubles((long) inputSize*outputSize, -1, 1).toArray());
        biases = new SimpleMatrix(1, outputSize, true,
                random.doubles(outputSize, -1, 1).toArray());
    }

    /**
     * Updates input size when previous layer has newly added neurons.
     * @param n amount of new neurons in previous layer
     */
    public void updateInputSize(int n) {
        Random random = new Random();

        // add new weights
        SimpleMatrix newWeights = new SimpleMatrix(this.weights.numRows() + n, this.weights.numCols());
        for (int i = 0; i < this.weights.numRows(); i++) {
            for (int j = 0; j < this.weights.numCols(); j++) {
                newWeights.set(i, j, this.weights.get(i, j));
            }
        }
        for (int i = 0; i < newWeights.getNumElements(); i++) {
            if (newWeights.get(i) == 0) {
                newWeights.set(i, random.nextDouble(-1, 1));
            }
        }
        this.weights = newWeights;
    }

    /**
     * Adds new neurons at the end of the layer
     * @param n amount how many new neurons should be added
     */
    public void addNeuron(int n) {
        Random random = new Random();

        // add new weights
        SimpleMatrix newWeights = new SimpleMatrix(this.weights.numRows(), this.weights.numCols() + n);
        for (int i = 0; i < this.weights.numRows(); i++) {
            for (int j = 0; j < this.weights.numCols(); j++) {
                newWeights.set(i, j, this.weights.get(i, j));
            }
        }
        for (int i = 0; i < newWeights.getNumElements(); i++) {
            if (newWeights.get(i) == 0) {
                newWeights.set(i, random.nextDouble(-1, 1));
            }
        }
        this.weights = newWeights;

        // add new biases
        SimpleMatrix newBiases = new SimpleMatrix(1, this.biases.numCols() + n);
        double[] newBiasValues = random.doubles(n, -1, 1).toArray();
        System.arraycopy(this.biases.getDDRM().data, 0, newBiases.getDDRM().data, 0, this.biases.numCols());
        System.arraycopy(newBiasValues, 0, newBiases.getDDRM().data, this.biases.numCols(), n);
        this.biases = newBiases;
    }

    @Override
    public SimpleMatrix forwardPropagation(SimpleMatrix inputs) {
        this.input = inputs;
        this.output = this.input.mult(this.weights).plus(this.biases);
        return this.output;
    }

    @Override
    public SimpleMatrix backwardPropagation(SimpleMatrix outputError, double learningRate) {
        SimpleMatrix inputError = outputError.mult(this.weights.transpose());
        SimpleMatrix weightsError = this.input.transpose().mult(outputError);

        this.weights = this.weights.plus(learningRate, weightsError);
        this.biases = this.biases.plus(learningRate, outputError);
        return inputError;
    }
}
